import anndata
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as stats
import seaborn as sns
import sqlalchemy as sa

import spherpro.bromodules.filters as filters
import spherpro.bromodules.plot_base as pltbase
import spherpro.configuration as conf
import spherpro.db as db

"""
TODO: move this to project specific package.
"""


class VariableBaseHelper:
    COL_CHANNELNAME = db.ref_planes.channel_name.key
    COL_CONDID = db.conditions.condition_id.key
    COL_CONDLEVEL = COL_CONDID + "level"
    COL_CONDNAME = db.conditions.condition_name.key
    COL_FILTERVAL = db.object_filters.filter_value.key
    COL_GOODNAME = "goodname"
    COL_IMGID = db.images.image_id.key
    COL_IMGLEVEL = COL_IMGID + "level"
    COL_IMID = db.images.image_id.key
    COL_ISNB = "isnb"
    COL_MEASID = db.measurements.measurement_id.key
    COL_MEASID = db.measurements.measurement_id.key
    COL_MEASNAME = db.measurement_names.measurement_name.key
    COL_MEASTYPE = db.measurements.measurement_type.key
    COL_OBJID = db.objects.object_id.key
    COL_SITEID = db.sites.site_id.key
    COL_SITELEVEL = COL_SITEID + "level"
    COL_VALUE = db.object_measurements.value.key
    COL_WORKING = "working"
    COL_D2RIM = "distrim"
    COL_PLATEID = db.conditions.plate_id.key
    COL_PLATELEVEL = COL_PLATEID + "level"


V = VariableBaseHelper


def cur_transf(x):
    return np.log10(x + 0.1)


class HelperVZ(pltbase.BasePlot):
    def __init__(self, bro):
        super().__init__(bro)
        self.filters = filters.Filters(bro)

    def get_pannelcsv(self):
        dat_pannelcsv = self.bro.data.pannel.drop_duplicates(subset="metal")
        return dat_pannelcsv

    def get_measuremeta(
        self,
        dat_pannelcsv,
        measurement_names=None,
        additional_measfilt=None,
        stack_names=None,
    ):
        if measurement_names is None:
            measurement_names = ["MeanIntensityComp", "NbMeanMeanIntensityComp"]
        if stack_names is None:
            stack_names = ["FullStackFiltered"]

        fil_measurements = sa.or_(
            sa.and_(
                db.stacks.stack_name.in_(stack_names),
                db.measurements.measurement_name.in_(measurement_names),
            ),
            sa.and_(
                db.stacks.stack_name == "ObjectStack",
                db.measurements.measurement_name == "dist-rim",
                db.ref_planes.channel_name == "object",
            ),
        )
        if additional_measfilt is not None:
            fil_measurements = sa.or_(fil_measurements, additional_measfilt)
        q = (
            self.bro.session.query(
                db.measurements,
                db.ref_planes.channel_name,
                db.stacks.stack_name,
                db.pannel,
            )
            .join(db.planes, db.measurements.plane_id == db.planes.plane_id)
            .join(db.stacks)
            .join(db.ref_planes)
            .filter(fil_measurements)
            .outerjoin(db.pannel, db.pannel.metal == db.ref_planes.channel_name)
        )
        dat_measmeta = self.bro.doquery(q)

        dat_measmeta = dat_measmeta.merge(
            dat_pannelcsv.loc[:, [db.pannel.metal.key, V.COL_GOODNAME, V.COL_WORKING]],
            on=db.pannel.metal.key,
            how="left",
        )
        dat_measmeta[V.COL_WORKING] = dat_measmeta[V.COL_WORKING].fillna(1)
        dat_measmeta[V.COL_ISNB] = dat_measmeta[V.COL_MEASNAME].map(
            lambda x: "Nb" if x.startswith("Nb") else "Int"
        )
        fil = dat_measmeta[V.COL_GOODNAME].isnull()
        dat_measmeta.loc[fil, V.COL_GOODNAME] = dat_measmeta.loc[fil, V.COL_CHANNELNAME]
        # dat_measmeta = dat_measmeta.set_index(V.COL_MEASID, drop=False)
        return dat_measmeta

    def get_imgmeta(self):
        q = (
            self.bro.session.query(
                db.images, db.conditions.condition_name, db.sites.site_id
            )
            .join(db.conditions)
            .join(db.acquisitions)
            .join(db.sites)
            .join(db.valid_images)
        )
        dat_imgmeta = self.bro.doquery(q)
        dat_imgmeta[V.COL_SITELEVEL] = get_level(dat_imgmeta, V.COL_SITEID)
        dat_imgmeta[V.COL_CONDLEVEL] = get_level(dat_imgmeta, V.COL_CONDID)
        dat_imgmeta[V.COL_IMGLEVEL] = get_level(dat_imgmeta, V.COL_IMGID)
        return dat_imgmeta

    def get_condmeta(self):
        dat = self.bro.doquery(self.bro.session.query(db.conditions))
        dat[V.COL_PLATELEVEL] = get_level(dat, V.COL_PLATEID)
        return dat

    def get_data(
        self,
        curcond=None,
        fil_good_meas=None,
        cond_ids=None,
        meas_ids=None,
        object_type=None,
        session=None,
        img_ids=True,
        obj_filter_query=None,
        legacy=True,
    ):

        q_obj = self.data.get_objectmeta_query(session=session)
        if img_ids:
            q_obj = q_obj.add_columns(db.objects.image_id)
        if (curcond is not None) or (cond_ids is not None):
            q_obj = q_obj.join(
                db.conditions, db.images.condition_id == db.conditions.condition_id
            )
            if curcond is not None:
                q_obj = q_obj.filter(db.conditions.condition_name == curcond)

            if cond_ids is not None:
                q_obj = q_obj.filter(db.conditions.condition_id.in_(cond_ids))
        if object_type is not None:
            q_obj = q_obj.filter(db.objects.object_type == object_type)

        if obj_filter_query is not None:
            q_obj = q_obj.filter(db.objects.object_id == obj_filter_query.c.object_id)
        q_meta = self.data.get_measmeta_query(session=session).with_entities(
            db.measurements.measurement_id, db.ref_stacks.scale
        )
        if fil_good_meas is not None:
            q_meta = q_meta.filter(fil_good_meas)
        if meas_ids is not None:
            q_meta = q_meta.filter(db.measurements.measurement_id.in_(meas_ids))
        dat_meas = self.bro.doquery(q_meta)
        dat_obj = self.bro.doquery(q_obj)

        dat = self.bro.io.objmeasurements.get_measurements(dat_obj, dat_meas)
        dat = self.bro.io.objmeasurements.scale_anndata(dat)

        if legacy:
            d = pd.DataFrame(
                dat.X, index=dat.obs.object_id, columns=dat.var.measurement_id
            ).stack()
            d.name = V.COL_VALUE
            dat = d.reset_index().merge(dat_obj)
        return dat

    def get_d2rim(self):
        measdict = self.bro.data.conf[conf.QUERY_DEFAULTS][conf.CORRDIST]
        obj = self.bro.data.conf[conf.QUERY_DEFAULTS][conf.OBJECTTYPE]
        fil = self.filters.measurements.get_measmeta_filter_statements(
            channel_names=[measdict[conf.DEFAULT_CHANNEL_NAME]],
            stack_names=[measdict[conf.DEFAULT_STACK_NAME]],
            measurement_names=[measdict[conf.DEFAULT_MEASUREMENT_NAME]],
            measurement_types=[measdict[conf.DEFAULT_MEASUREMENT_TYPE]],
        )
        dat = self.get_data(fil_good_meas=fil, object_type=obj)
        dat = rename_measurement(dat, V.COL_D2RIM)
        return dat

    def get_object_meta(self, object_types=None):
        if object_types is None:
            obj = self.bro.data.conf[conf.QUERY_DEFAULTS][conf.OBJECTTYPE]
            object_types = [obj]
        q = self.bro.session.query(db.objects)
        if len(object_types) > 0:
            q = q.filter(db.objects.object_type.in_(object_types))
        return self.bro.doquery(q)

    def get_fildats(self, filnames, outnames=None):
        d = self.bro.doquery(
            self.bro.session.query(
                db.object_filters.object_id,
                db.object_filters.filter_value,
                db.object_filter_names.object_filter_name,
            )
            .join(db.object_filter_names)
            .filter(db.object_filter_names.object_filter_name.in_(filnames))
        )
        d = (
            d.assign(
                **{
                    db.object_filters.filter_value.key: lambda x: x[
                        db.object_filters.filter_value.key
                    ]
                }
            )
            .pivot(
                values=db.object_filters.filter_value.key,
                index=db.object_filters.object_id.key,
                columns=db.object_filter_names.object_filter_name.key,
            )
            .reset_index()
        )
        if outnames is not None:
            d = d.rename({f: o for f, o in zip(filnames, outnames)}, axis=1)
        return d

    def get_sitemeta(self):
        q = (
            self.bro.session.query(
                db.sites, db.acquisitions, db.slideacs, db.slides, db.sampleblocks
            )
            .join(db.acquisitions)
            .join(db.slideacs)
            .join(db.slides)
            .join(db.sampleblocks)
        )
        dat_sitemeta = self.bro.doquery(q)
        dat_sitemeta = dat_sitemeta.loc[:, ~dat_sitemeta.columns.duplicated()]
        return dat_sitemeta

    def plt_clustmatp(self, dat_meas_raw, dat_measmeta):
        dat_meas = dat_meas_raw.pivot_table(
            values=V.COL_VALUES, index=V.COL_OBJID, columns=V.COL_MEASID
        )

        corrdat, pval = stats.spearmanr(dat_meas.dropna())
        colnames = dat_measmeta.loc[dat_meas.columns.values, :].apply(
            lambda x: " - ".join([x[V.COL_ISNB], x[V.COL_GOODNAME], x[V.COL_CHANNELN]]),
            axis=1,
        )

        cg = sns.clustermap(pd.DataFrame(corrdat, index=colnames, columns=colnames))
        x = plt.setp(cg.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)
        x = plt.setp(cg.ax_heatmap.xaxis.get_majorticklabels(), rotation=90)
        cg.fig.subplots_adjust(bottom=0.3)
        cg.fig.subplots_adjust(right=0.7)
        return cg

    def transf_intensities(self, dat, dat_measmeta):
        V.COL_VALUE = db.object_measurements.value.key
        ids = dat_measmeta.loc[
            dat_measmeta[V.COL_MEASTYPE] == "Intensity", V.COL_MEASID
        ]
        fil = dat[V.COL_MEASID].isin(ids)
        dat.loc[fil, V.COL_VALUE] = cur_transf(dat.loc[fil, V.COL_VALUE].values)
        return dat

    def plt_clustmatp_pearson(self, dat_meas_raw, dat_measmeta):
        dat_meas = dat_meas_raw.copy()
        dat_meas = self.transf_intensities(dat_meas, dat_measmeta)
        dat_meas = dat_meas.pivot_table(
            values=V.COL_VALUES, index=V.COL_OBJID, columns=V.COL_MEASID
        )
        dat_meas = dat_meas.dropna()
        corrdat = np.corrcoef(dat_meas.T)
        colnames = dat_measmeta.loc[dat_meas.columns.values, :].apply(
            lambda x: " - ".join([x[V.COL_ISNB], x[V.COL_GOODNAME], x[V.COL_CHANNELN]]),
            axis=1,
        )
        cg = sns.clustermap(pd.DataFrame(corrdat, index=colnames, columns=colnames))
        x = plt.setp(cg.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)
        x = plt.setp(cg.ax_heatmap.xaxis.get_majorticklabels(), rotation=90)
        cg.fig.subplots_adjust(bottom=0.3)
        cg.fig.subplots_adjust(right=0.7)
        return cg

    def get_fil_good_meas(self, dat_measmeta, col_working=V.COL_WORKING):
        fil_good_meas = db.measurements.measurement_id.in_(
            [
                int(i)
                for i in dat_measmeta.loc[dat_measmeta[col_working] == 1, V.COL_MEASID]
            ]
        )
        return fil_good_meas

    def get_imgs_for_cond(self, condition_name):
        q = (
            self.bro.session.query(db.images.image_id)
            .join(db.conditions)
            .join(db.valid_images)
            .filter(db.conditions.condition_name == condition_name)
        )
        return [i[0] for i in q.all()]

    # Things for anndata export
    def get_full_obj_meta(self):
        dat_objmeta = self.get_object_meta()
        q = (
            self.bro.session.query(
                db.images,
                db.conditions,
                db.acquisitions,
                db.sites,
                db.slideacs,
                db.slides,
            )
            .join(db.conditions)
            .join(db.acquisitions)
            .join(db.sites)
            .join(db.slideacs)
            .join(db.slides)
            .join(db.valid_images)
        )
        dat_condition = self.bro.doquery(q)
        self.bro.data._read_experiment_layout()
        dat_d2rim = self.get_d2rim()
        dat_full_objmeta = (
            dat_condition.merge(self.bro.data.experiment_layout)
            .pipe(remove_duplicated_columns)
            .merge(dat_objmeta)
            .merge(dat_d2rim)
            .set_index(db.objects.object_id.key, drop=True)
        )
        return dat_full_objmeta

    def get_full_meas_meta(self):
        dat_measmeta = self.bro.doquery(
            self.bro.data.get_measmeta_query().add_columns(
                db.stacks.stack_name, db.ref_planes.channel_name
            )
        )
        dat_full_measmeta = (
            dat_measmeta.merge(
                self.bro.data.pannel,
                left_on=db.ref_planes.channel_name.key,
                right_on="metal",
                how="left",
            )
            .pipe(remove_duplicated_columns)
            .set_index(db.measurements.measurement_id.key, drop=True)
        )
        return dat_full_measmeta

    def convert_to_anndata(self, dat_meas, dat_obs, dat_var):
        dat_meas = dat_meas.pivot_table(
            values=db.object_measurements.value.key,
            index=db.objects.object_id.key,
            columns=db.measurements.measurement_id.key,
        )
        dat = anndata.AnnData(
            dat_meas,
            obs=dat_obs.loc[dat_meas.index, :],
            var=dat_var.loc[dat_meas.columns, :],
        )
        return dat


def get_level(dat, idcol):
    return dat[idcol].map(lambda x: f"{idcol}_{int(x)}")


def rename_measurement(dat, name):
    dat = dat.rename({V.COL_VALUE: name}, axis=1)
    dat = dat.drop(V.COL_MEASID, axis=1)
    return dat


def remove_duplicated_columns(dat):
    return dat.loc[:, ~dat.columns.duplicated()]


class Renamer(object):
    """
    A class to rename & unrename integer varaible names to make them compatible with the linear model
    """

    def __init__(self):
        self.d = dict()

    def rename(self, x):
        rx = "c" + str(x)
        self.d.update({rx: x})
        return rx

    def unrename(self, x):
        return self.d.get(x, x)



import colorcet
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import pandas as pd

import spherpro.bromodules.plot_base as plot_base
import spherpro.db as db

LABEL_CBAR = "# of all cells with valid barcodes"
LABEL_Y = "# of cells with\nmost prominent barcode"
LABEL_X = "# of cells with second most prominent barcode"
PLT_TITLE = "Debarcoding Quality"
CBAR_HUE = db.images.bc_valid.key


class PlotDebarcodeCells(plot_base.BasePlot):
    def __init__(self, bro):
        super().__init__(bro)

    def plot_debarcoded_cells(
        self,
        img_id,
        color_invalid="#F8F8F8",
        base_colormap=colorcet.glasbey,
        colorbar=False,
        ax=None,
        title=None,
        **kwargs,
    ):
        # get the conditions of the block
        blockid = (
            self.session.query(db.sampleblocks).join(db.conditions).join(db.images)
        ).subquery()
        unicols = [
            c[0]
            for c in self.session.query(db.conditions.condition_id)
            .filter(db.conditions.sampleblock_id == blockid.c.sampleblock_id)
            .all()
        ]
        cmap = [color_invalid] + base_colormap
        ncol = max(unicols) + 1
        mymap = mcolors.LinearSegmentedColormap.from_list("my_colormap", cmap, N=ncol)

        if title is None:
            title = f"ImgId: {img_id}"
        # TODO: get barcoding variable names from the debarcoding module.
        return self.bro.plots.heatmask.plt_heatplot(
            [img_id],
            "barcode",
            "ObjectStack",
            "object",
            title=title,
            transform=None,
            colorbar=colorbar,
            cmap=mymap,
            ax=ax,
            crange=[0, ncol - 1],
            **kwargs,
        )


class PlotDebarcodeQuality(plot_base.BasePlot):
    def __init__(self, bro):
        super().__init__(bro)
        # make the dependency explicit

    def quality_plot(self, filename=None, show=None, cm=None):
        """
        Plots a quality Plot for the debarcoding. This function requires
        a debarcoded dataset!

        Args:
            str filename: if specified, saves the plot to the location
            bool show: should plt.show be executed? default False
            cm: color map to use
        Returns:
            plt, ax from plot
        """
        if show is None:
            show = False
        if not self.bro.is_debarcoded:
            raise NameError("Please use a debarcoded dataset or debarcode this one!")

        # get data
        table, zeros = self._get_data()
        # plot data
        plt, ax = self._produce_plot(table, zeros, cm=cm)

        if filename is not None:
            plt.savefig(filename)
        else:
            if show:
                plt.show()

        return plt, ax

    def _get_data(self):
        q = self.data.main_session.query(db.images)
        zeros = q.filter(db.images.condition_id == None).count()
        q = q.filter(db.images.condition_id.isnot(None)).statement
        table = pd.read_sql_query(q, self.data.db_conn)
        return table, zeros

    def _produce_plot(self, data, zeros, cm=None):
        fig, ax = plt.subplots()

        if cm is None:
            cm = plt.cm.get_cmap("winter")

        y = data[db.images.bc_highest_count.key]
        x = data[db.images.bc_second_count.key]
        sc = ax.scatter(x, y, alpha=0.7, edgecolors="none", c=data[CBAR_HUE], cmap=cm)
        upper = max(x.max() * 1.1, y.max() * 1.1)
        ax.set_xlim([0, upper])
        ax.set_ylim([0, upper])
        # ax.legend()
        ax.grid(True)
        ax.set_aspect(1)
        plt.plot(
            [0, upper],
            [0, upper],
            "k-",
            c="grey",
            lw=1,
            alpha=0.5,
            label="_not in legend",
        )
        cbar = plt.colorbar(sc)
        cbar.set_label(LABEL_CBAR)

        plt.ylabel(LABEL_Y)
        plt.xlabel(LABEL_X)

        plt.title(PLT_TITLE)
        return plt, ax



import copy
from typing import Iterable, List, Optional

import ipywidgets as ipw
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import pycytools.library
import pycytools as pct

import spherpro.bromodules.plot_base as plot_base
import spherpro.db as db


# TODO: move to the pycytools!
def logtransf_data(x):
    xmin = min(x[x > 0])
    return np.log(x + xmin)


def asinhtransf_data(x):
    return np.arcsinh(x / 5)


transf_dict = {
    "none": lambda x: x,
    "log": logtransf_data,
    "asinh": asinhtransf_data,
    "sqrt": np.sqrt,
}

COL_OBJECT_TYPE = db.objects.object_type.key


class InteractiveHeatplot(object):
    def __init__(self, session, plotter):
        self.session = session
        self.plotter = plotter

    def selector_basic(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        name_dict = {
            m: n for m, n in self.session.query(db.pannel.metal, db.pannel.target).all()
        }
        # make the sites widget
        sites = [s[0] for s in self.session.query(db.sites.site_id).all()]
        # add the imageidx widget and make it adapt to the
        channel_names = [
            q[0] for q in self.session.query(db.ref_planes.channel_name).all()
        ]
        stack_names = [q[0] for q in self.session.query(db.stacks.stack_name).all()]
        object_names = [
            q[0] for q in self.session.query(db.masks.object_type).distinct()
        ]
        measurement_names = [
            q[0]
            for q in self.session.query(db.measurement_names.measurement_name).all()
        ]
        channel_names_info = [
            pct.library.name_from_metal(c, name_dict) for c in channel_names
        ]

        ipw.interact(
            self._selector_basic_plot,
            site=sites,
            roi_idx=ipw.IntSlider(min=0, max=30, continuous_update=False),
            img_idx=ipw.IntSlider(min=0, max=70, continuous_update=False),
            stat=measurement_names,
            stack=stack_names,
            channel=channel_names_info,
            transform=list(transf_dict.keys()),
            censor_min=ipw.FloatSlider(
                min=0, max=0.5, value=0, step=0.0001, continuous_update=False
            ),
            censor_max=ipw.FloatSlider(
                min=0.5, max=1, value=1, step=0.0001, continuous_update=False
            ),
            keepRange=ipw.Checkbox(),
            filter_hq=ipw.Checkbox(value=True),
            ax=ipw.fixed(ax),
        )

    def _selector_basic_plot(
        self,
        site,
        roi_idx,
        img_idx,
        stat,
        stack,
        channel,
        transform,
        censor_min,
        censor_max,
        keepRange,
        filter_hq,
        ax,
    ):
        q = (
            self.session.query(db.images.image_id)
            .join(db.valid_images)
            .join(db.acquisitions)
            .filter(db.acquisitions.site_id == site)
        )
        if roi_idx > 0:
            r_id = (
                self.session.query(db.acquisitions.acquisition_id)
                .filter(db.acquisitions.site_id == site)
                .order_by(db.acquisitions.acquisition_id)
                .offset(roi_idx - 1)
            ).first()
            if r_id is None:
                return
            q = q.filter(db.acquisitions.acquisition_id == r_id)

        if img_idx == 0:
            imnr = [r[0] for r in q.distinct()]
        else:
            imnr = [q.order_by(db.images.image_id).offset(img_idx - 1).first()[0]]
            print(imnr)

        if imnr[0] is None:
            return
        metal = pct.library.metal_from_name(channel)
        self.plotter.plt_heatplot(
            imnr,
            stat,
            stack,
            metal,
            transform=transform,
            censor_min=censor_min,
            censor_max=censor_max,
            keepRange=keepRange,
            ax=ax,
            title=channel,
        )

    def get_dynamic_selector(self, plotfkt):
        ALL = "all"
        name_dict = {
            m: n for m, n in self.session.query(db.pannel.metal, db.pannel.target).all()
        }
        # make the sites widget
        sites = [s[0] for s in self.session.query(db.sites.site_id).all()]
        w_sites = ipw.Select(sites)
        # add the roi widget and make it adapt depending on the chosen site
        w_roi = ipw.Select([ALL] + self.get_roi_ids(w_sites.value))

        def update_roi_range(*args):
            w_roi.options = [ALL] + self.get_roi_ids(w_sites.value)

        w_sites.observe(update_roi_range, "value")
        # add the imageidx widget make it adapt depending on the chosen site
        w_image = ipw.Select([ALL] + self.get_image_ids(w_sites.value))

        def update_img_range(*args):
            w_image.options = [ALL] + self.get_img_ids(w_roi.value)

        w_roi.observe(update_img_range, "value")
        w_imgidx = ipw.__version__
        channel_names = [
            q[0] for q in self.data.main_session.query(db.ref_planes.channel_name).all()
        ]
        stack_names = [
            q[0] for q in self.data.main_session.query(db.stacks.stack_name).all()
        ]
        object_names = [
            q[0] for q in self.data.main_session.query(db.masks.object_type).distinct()
        ]
        measurement_names = [
            q[0]
            for q in self.data.main_session.query(
                db.measurement_names.measurement_name
            ).all()
        ]
        channel_names_info = [
            pct.library.name_from_metal(c, name_dict) for c in channel_names
        ]

        ipw.interact(
            self.plt_heatplot,
            site=sites,
            roi_idx=ipw.IntSlider(min=0, max=30, continuous_update=False),
            img_idx=ipw.IntSlider(min=0, max=30, continuous_update=False),
            stat=measurement_names,
            stack=stack_names,
            channel=channel_names_info,
            transform=list(transf_dict.keys()),
            censor_min=ipw.FloatSlider(
                min=0, max=0.5, value=0, step=0.0001, continuous_update=False
            ),
            censor_max=ipw.FloatSlider(
                min=0.5, max=1, value=1, step=0.0001, continuous_update=False
            ),
            keepRange=ipw.Checkbox(),
            filter_hq=ipw.Checkbox(value=True),
            ax=ipw.fixed(ax),
        )

    def get_roi_ids(self, site_id):
        q_rois = (
            self.session.query(db.acquisitions.acquisition_id)
            .filter(db.acquisitions.site_id == site_id)
            .order_by(db.acquisitions.acquisition_id)
            .all()
        )
        rois = [r[0] for r in q_rois]
        return rois

    def get_img_ids(self, site_id):
        q_rois = (
            self.session.query(db.acquisitions.acquisition_id)
            .filter(db.acquisitions.site_id == site_id)
            .order_by(db.acquisitions.acquisition_id)
            .all()
        )
        rois = [r[0] for r in q_rois]
        return rois


class PlotHeatmask(plot_base.BasePlot):
    def __init__(self, bro):
        super().__init__(bro)
        self.io_masks = self.bro.io.masks
        self.filter_measurements = self.bro.filters.measurements
        self.objmeasurements = self.bro.io.objmeasurements
        # TODO: add defaults from config
        self.measure_idx = [  # idx_name, default
            (db.ref_planes.channel_name.key, None),
            (db.stacks.stack_name.key, "FullStack"),
            (db.measurement_names.measurement_name.key, "MeanIntensity"),
            (db.measurement_types.measurement_type.key, None),
        ]
        self.interactive = InteractiveHeatplot(self.data.main_session, self)

    def _prepare_masks(
        self, image_ids: Iterable[int], object_type: str
    ) -> List[np.ndarray]:
        """
        Gets masks for the images
        Args:
            image_ids: image_ids
            object_type: object_type

        Returns: List of masks.

        """
        masks = [self.io_masks.get_mask(i, object_type) for i in image_ids]
        return masks

    def _prepare_slices(self, image_numbers):
        dat = (
            self.session.query(
                db.images.image_id,
                db.images.image_pos_x,
                db.images.image_pos_y,
                db.images.image_shape_w,
                db.images.image_shape_h,
            ).filter(db.images.image_id.in_(image_numbers))
        ).all()
        slice_dict = {
            i: (np.s_[x : (x + w)], np.s_[y : (y + h)]) for i, x, y, h, w in dat
        }
        slices = [slice_dict[i] for i in image_numbers]
        return slices

    def get_heatmask_data(
        self,
        measurement_dict,
        image_ids=None,
        filters=None,
        valid_objects=True,
        valid_images=True,
        object_type="cell",
    ):

        if filters is None:
            filters = []

        filter_statement = self.filter_measurements.get_measmeta_filter_statements(
            *[[measurement_dict.get(o, d)] for o, d in self.measure_idx]
        )

        q_meas = (
            self.data.get_measmeta_query(session=self.session)
            .filter(filter_statement)
            .add_column(db.ref_stacks.scale)
        )
        q_obj = self.data.get_objectmeta_query(
            session=self.session, valid_objects=valid_objects, valid_images=valid_images
        ).filter(db.objects.object_type == object_type)

        # add more output columns
        q_obj = q_obj.add_columns(db.objects.object_number)

        if image_ids is not None:
            q_obj = q_obj.filter(db.images.image_id.in_(image_ids))
        if len(filters) > 0:
            q_obj = q_obj.join(db.object_filters)
        for fil in filters:
            # TODO: this NEEDs to be fixed as it wont work as expected with multiple filters!
            # This needs to be done with subqueries!
            q_obj = q_obj.filter(fil)

        data = self.objmeasurements.get_measurements(q_obj=q_obj, q_meas=q_meas)
        data = self.objmeasurements.scale_anndata(data)
        data = self.objmeasurements.convert_anndata_legacy(data)
        return data

    def assemble_heatmap_image(
        self,
        dat_cells: pd.DataFrame,
        image_ids: Optional[Iterable[int]] = None,
        cut_slices: Optional[Iterable[type(np.s_)]] = None,
        cut_masks: Optional[Iterable[np.ndarray]] = None,
        out_shape: Optional[Iterable[int]] = None,
        value_var: Optional[str] = None,
        object_type: Optional[str] = None,
    ):

        """
        Generates an image with the data mapped on the mask and the masks assembled into the original image.

        Args:
            dat_cells: A dataframe with the data to be mapped, most contain the value_var and object_number columns
            image_ids: A list of image ids
            cut_slices: [optional] A list of np.s_ slice -> indicates where the different cuts are located in the big image
            cut_masks: [optional] A list of image masks with labels corresponding to object_numbers
            out_shape: [optional] Final size of the assembled image -> Calculated if not provided
            value_var: [optional] The variable containing the value to be mapped
            object_type: [optional] The object type -> Required to get the correct mask if 'object_type' is not a column
                        in the data.

        Returns:
            An masked array. NAN values indicate non-mapped values. The mask covers regions that were not segmented.

        """

        cut_id_name = db.images.image_id.key
        cell_id_name = db.objects.object_number.key

        if value_var is None:
            value_var = db.object_measurements.value.key

        if image_ids is None:
            image_ids = dat_cells[cut_id_name].unique().tolist()

        if cut_slices is None:
            cut_slices = self._prepare_slices(image_ids)

        if cut_masks is None:
            if object_type is None:
                assert (
                    COL_OBJECT_TYPE in dat_cells.columns
                ), f"{COL_OBJECT_TYPE} column not in data."
                assert dat_cells[COL_OBJECT_TYPE].nunique() == 1, (
                    f"Multiple {COL_OBJECT_TYPE}s present in data."
                    "cannot choose mask type."
                )
                object_type = dat_cells[COL_OBJECT_TYPE].iloc[0]
            cut_masks = self._prepare_masks(image_ids, object_type)

        if out_shape is None:
            x_start, x_stop, y_start, y_stop = zip(
                *[(slx.start, slx.stop, sly.start, sly.stop) for slx, sly in cut_slices]
            )
            new_shape = (max(x_stop), max(y_stop))
        else:
            new_shape = out_shape

        pimg = np.empty(new_shape)
        pimg[:] = np.NAN
        notbg = np.zeros(new_shape)

        dat_cells = dat_cells.set_index(cut_id_name)

        for cid, sl, mask in zip(image_ids, cut_slices, cut_masks):
            if cid in dat_cells.index:
                intensity = dat_cells.loc[cid, :]
                intensity = intensity.set_index(cell_id_name)
                intensity = intensity[value_var]
                timg = pct.library.map_series_on_mask(
                    mask, intensity, label=intensity.index
                )
                pimg_sl = pimg[sl]

                fil = (
                    (timg.mask == False)
                    & (np.isnan(timg.imag) == False)
                    & (np.isnan(pimg_sl) == True)
                )
                pimg_sl[fil] = timg[fil]
                fil2 = notbg[sl] == 0
                notbg[sl][fil2] = (timg.mask == False)[fil2]

        pimg = np.ma.array(pimg, mask=(notbg == 0))

        if out_shape is None:
            pimg = pimg[min(x_start) :, min(y_start) :]

        return pimg

    @staticmethod
    def do_heatplot(
        img,
        title=None,
        crange=None,
        ax=None,
        update_axrange=True,
        cmap=None,
        colorbar=True,
        cmap_mask=None,
        cmap_mask_alpha=0.3,
        bad_color="k",
        bad_alpha=1,
    ):

        # if crange is None:
        #    crange=(np.nanmin(img[:]), np.nanmax(img[:]))

        if cmap is None:
            cmap = plt.cm.viridis
        cmap = copy.copy(cmap)
        cmap.set_bad(bad_color, bad_alpha)
        if ax is None:
            plt.close()
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        if len(ax.images) == 0:
            cax = ax.imshow(img, cmap=cmap, interpolation="nearest")
            if colorbar:
                fig.colorbar(cax)
        else:
            cax = ax.images[0]
            if len(ax.images) == 2:
                cax_mask = ax.images[1]
                cax_mask.remove()

        cax.set_data(np.flipud(img))
        cax.set_extent([0, img.shape[1], 0, img.shape[0]])
        if crange is not None:
            cax.set_clim(crange[0], crange[1])

        # update bounts

        if update_axrange == True:
            ax.set_xbound(lower=0, upper=img.shape[1])
            ax.set_ybound(lower=0, upper=img.shape[0])
        else:
            bndx = ax.get_xbound()
            ax.set_xbound(upper=min(img.shape[1], bndx[1]))
            bndy = ax.get_ybound()
            ax.set_ybound(upper=min(img.shape[0], bndy[1]))

        if title is not None:
            ax.set_title(title)
        # fig.colorbar(cax)
        # slightly color the non background but not colored cells
        if hasattr(img, "mask"):
            mask_img = np.isnan(img)
            if np.any(mask_img):
                mask_img = np.ma.array(mask_img, mask=img.mask | (mask_img == False))
                if cmap_mask is None:
                    cmap_mask = "Greys"
                ax.imshow(
                    mask_img == 1,
                    cmap=cmap_mask,
                    alpha=cmap_mask_alpha,
                    interpolation="nearest",
                )
        return ax

    def plt_heatplot(
        self,
        img_ids,
        stat,
        stack,
        channel,
        transform=None,
        censor_min=0,
        censor_max=1,
        keepRange=False,
        filters=None,
        filter_hq=None,
        ax=None,
        title=None,
        colorbar=True,
        transform_fkt=None,
        cmap=None,
        crange=None,
        valid_objects=True,
        valid_images=True,
        **kwargs,
    ):
        """
        Retrieves images form the database and maps then on masks
        TODO: update
        Args:
            img_ids:
            stat: The KEY_MEASUREMENTNAME
            stack: the Stackname
            channel: the ChannelName
            transform: a transform
            ...
        Returns:
            The axis with the heatplot
        """

        # if filter_hq:
        #     fil = [sa.and_(db.object_filters.=='is-hq', db.object_filters.filter_value==True)]
        # else:
        #     fil = None
        if transform is None:
            transform = "none"

        fil = filters
        # print('Start loading...')
        data = self.get_heatmask_data(
            {
                db.ref_planes.channel_name.key: channel,
                db.stacks.stack_name.key: stack,
                db.measurement_names.measurement_name.key: stat,
            },
            image_ids=img_ids,
            filters=fil,
            valid_images=valid_images,
            valid_objects=valid_objects,
        )
        # print(data.shape)
        # print('Finished loading!')
        col_val = db.object_measurements.value.key
        if transform_fkt is None:
            transform_fkt = transf_dict[transform]
        data[col_val] = transform_fkt(data[col_val])
        if data.shape[0] == 0:
            if ax is None:
                fig = plt.figure()
                a = fig.axes[0]
            else:
                a = ax
        else:
            img = self.assemble_heatmap_image(data)
            if crange is not None:
                pass
            elif (censor_min > 0) | (censor_max < 1):
                crange = (
                    np.percentile(data[col_val], censor_min * 100),
                    np.percentile(data[col_val], censor_max * 100),
                )
            else:
                crange = (data[col_val].min(), data[col_val].max())

            if title is None:
                title = channel

            a = self.do_heatplot(
                img,
                title=title,
                crange=crange,
                ax=ax,
                update_axrange=~keepRange,
                colorbar=colorbar,
                cmap=cmap,
                **kwargs,
            )
        a.axis("off")
        return a



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()

setup(
    name='spherpro',
    version='0.1.3post1',
    description='Tool to analize tumor spheroid data',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Vito Zanotelli, Matthias Leutenegger, Bodenmiller Lab UZH',
    author_email='vito.zanotelli@uzh.ch',
    url='https://github.com/bodenmillerlab/spherpro',
    license='BSD-3 License',
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'anndata',
        'colorcet',
        'imctools==1.0.7',
        'matplotlib',
        'numpy',
        'pandas',
        'pyyaml',
        'requests',
        'scikit-image',
        'scikit-learn',
        'scipy',
        'seaborn',
        'sqlalchemy',
        'tifffile',
        'sphinx==3.2.1',
        'plotnine',
        'matplotlib_scalebar',
        'pycytools>0.6',
        'ipywidgets']
)



import pandas as pd

import spherpro.bromodules.plot_base as base
import spherpro.configuration as conf
import spherpro.db as db

MEAS_ID = db.measurements.measurement_id.key
MEAS_NAME = db.measurements.measurement_name.key
MEAS_TYPE = db.measurements.measurement_type.key
MEAS_PLANE = db.measurements.plane_id.key


class MeasurementMaker(base.BasePlot):
    def __init__(self, bro):
        super().__init__(bro)

    def register_measurement_name(self, new_measname):
        x = db.measurement_names()
        x.measurement_name = new_measname
        self.session.merge(x)
        self.session.commit()

    def register_measurement_type(self, new_meastype):
        x = db.measurement_types()
        x.measurement_type = new_meastype
        self.session.merge(x)
        self.session.commit()

    def register_objects(self, object_meta, assume_new=False):
        """
        Registers an object metadata table
        Needs to have the columns:
            object_type
            image_id
            object_number
        Registers them and returns a table
            object_type
            image_id
            object_number
            object_id
            image_id

        If assume_new is True, the objects are assumed to be guaranteed
        new - use with care!
        """
        COL_OBJ_ID = db.objects.object_id.key
        img_dict = {
            n: i
            for n, i in self.session.query(db.images.image_number, db.images.image_id)
        }
        object_meta[db.images.image_id.key] = object_meta[
            db.images.image_number.key
        ].replace(img_dict)
        if assume_new == False:
            object_meta[COL_OBJ_ID] = [
                l[0] if l is not None else None
                for l in [
                    self.session.query(COL_OBJ_ID)
                    .filter(
                        db.objects.object_number == row[db.objects.object_number.key],
                        db.objects.image_id == row[db.objects.image_id.key],
                        db.objects.object_type == row[db.objects.object_type.key],
                    )
                    .one_or_none()
                    for idx, row in object_meta.iterrows()
                ]
            ]
        else:
            object_meta[COL_OBJ_ID] = None
        fil = object_meta[COL_OBJ_ID].isnull()
        if sum(fil) > 0:
            object_meta.loc[fil, COL_OBJ_ID] = self.bro.data._query_new_ids(
                db.objects.object_id, sum(fil)
            )
            # object_meta[COL_OBJ_ID] = object_meta[COL_OBJ_ID].astype(np.int)
            self.bro.data._bulkinsert(object_meta.loc[fil, :], db.objects)
        return object_meta

    def register_single_measurement(self, measurement_name, measurement_type, plane_id):
        dat_measure_meta = pd.DataFrame(
            {
                db.measurements.measurement_name.key: [measurement_name],
                db.measurements.plane_id.key: [plane_id],
                db.measurements.measurement_type.key: [measurement_type],
            }
        )
        dat_measure_meta = self.register_measurements(dat_measure_meta)
        return dat_measure_meta[db.measurements.measurement_id.key].values[0]

    def register_measurements(self, measure_meta):
        """
        Retisters a table with measurments. The ID column
        will be added automatically.
        """
        measure_meta = measure_meta.copy()
        # make sure the names and types are present:
        for m in measure_meta[MEAS_NAME].unique():
            self.register_measurement_name(m)
        for m in measure_meta[MEAS_TYPE].unique():
            self.register_measurement_type(m)
        measure_meta[MEAS_ID] = [
            l[0] if l is not None else None
            for l in [
                self.session.query(db.measurements.measurement_id)
                .filter(
                    db.measurements.measurement_name == row[MEAS_NAME],
                    db.measurements.measurement_type == row[MEAS_TYPE],
                    db.measurements.plane_id == row[MEAS_PLANE],
                )
                .one_or_none()
                for idx, row in measure_meta.iterrows()
            ]
        ]

        fil = measure_meta[MEAS_ID].isnull()
        if sum(fil) > 0:
            measure_meta.loc[fil, MEAS_ID] = self.bro.data._query_new_ids(
                db.measurements.measurement_id, sum(fil)
            )
            # measure_meta[MEAS_ID] = measure_meta[MEAS_ID].astype(int)
            self.bro.data._bulkinsert(measure_meta.loc[fil, :], db.measurements)
        return measure_meta

    def add_object_measurements(self, dat_meas, replace=True, drop_all_old=False):
        self.bro.io.objmeasurements.add_anndata_datmeasurements(
            dat_meas, replace=replace, drop_all_old=drop_all_old
        )

    def delete_measurements_by_ids(self, meas_ids):
        q = self.bro.session.query(db.object_measurements).filter(
            db.object_measurements.measurement_id.in_([int(i) for i in meas_ids])
        )
        q.delete(synchronize_session=False)
        self.bro.session.commit()

    def get_object_plane_id(self):
        """
        Gets the object plane id, based on defaults.

        The object plane is a plane to capture object level measurements,
        not related to any channel, e.g. size/shape/location...
        """
        def_obj = self.bro.data.conf[conf.QUERY_DEFAULTS][conf.OBJECT_DEFAULTS]
        OUT_STACK = def_obj[conf.DEFAULT_STACK_NAME]
        OUT_CHANNEL_TYPE = def_obj[conf.DEFAULT_CHANNEL_TYPE]
        OUT_CHANNEL_NAME = def_obj[conf.DEFAULT_CHANNEL_NAME]
        plane_id = (
            self.bro.session.query(db.planes.plane_id)
            .join(db.stacks)
            .join(db.ref_stacks)
            .join(db.ref_planes)
            .filter(
                db.stacks.stack_name == OUT_STACK,
                db.ref_planes.channel_type == OUT_CHANNEL_TYPE,
                db.ref_planes.channel_name == OUT_CHANNEL_NAME,
            )
        ).one()[0]
        return plane_id



import collections
import copy

import yaml

import spherpro.db as db

"""
This module defines the configuration file structure,
defaults, required files, loading as well as validation functions.
"""

"""
List of keywords fields used in the configuration file
"""
QUERY_DEFAULTS = "query_defaults"
CHANNEL_MEASUREMENTS = "channel_measurements"
RAWDIST = "distance_measure_raw"
CORRDIST = "distance_measure_corrected"
OBJECT_DEFAULTS = "object_defaults"
DEFAULT_OBJECT_TYPE = db.objects.object_type.key
DEFAULT_MEASUREMENT_NAME = db.measurement_names.measurement_name.key
DEFAULT_STACK_NAME = db.stacks.stack_name.key
DEFAULT_MEASUREMENT_TYPE = db.measurement_types.measurement_type.key
DEFAULT_OBJ_TYPE = db.objects.object_type.key
DEFAULT_CHANNEL_NAME = db.ref_planes.channel_name.key
DEFAULT_CHANNEL_TYPE = db.ref_planes.channel_type.key

BACKEND = "backend"
BARCODE_CSV = "barcode_csv"
CHANNEL_NAME = "channel_name_col"
CHANNEL_TYPE = "channel_type"
CHANNEL_TYPE_DEFAULT = "IMC"
CONDITION = "condition_col"
CPOUTPUT = "cpoutput"
CP_DIR = "cp_dir"
DEFAULT_OBJECT = "default_object"
DISPLAY_NAME = "display_name_col"
FILENAME_COL = "filename_col"
FILETYPE = "filetype"
GROUP_SITE = "group_site"
ID = "id_col"
IMAGENUMBER = "image_number_col"
IMAGENUMBER_FROM = "first_image_number_col"
IMAGENUMBER_TO = "second_image_number_col"
IMAGES_CSV = "images_csv"
LAYOUT_CSV = "layout_csv"
MASK_FILENAME_PREFIX = "mask_filename_col_prefix"
STACKIMG_FILENAME_PREFIX = "stackimg_filename_col_prefix"
MEASUREMENT_CSV = "measurement_csv"
MODNAME = "modname_col"
MODPRE = "modpre_col"
NAME = "name_col"
OBJECTTYPE = "object_type"
OBJECTTYPE_FROM = "first_object_type_col"
OBJECTTYPE_TO = "second_object_type_col"
OBJECTNUMBER = "object_number_col"
OBJECTNUMBER_FROM = "first_object_number_col"
OBJECTNUMBER_TO = "second_object_number_col"
OBJECTS = "objects"
PANEL_CSV = "pannel_csv"
PANNEL_CSV = "pannel_csv"
PANNEL_ID = "pannel_id"
PARENT = "parent_col"
PATH = "path"
PLATE = "plate_col"
REF = "ref_col"
RELATIONSHIP = "relationship"
RELATION_CSV = "relation_csv"
SCALING_PREFIX = "scaling_prefix"
SEP = "sep"
STACK = "stack_col"
STACK_DIR = "stack_dir"
STACK_RELATIONS = "stack_relations"
TYPE = "type_col"
WELL_COL = "well_col"
MASK_DIR = "mask_dir"
STACKIMG_DIR = "stackimg_dir"
GROUP_SITE = "group_site"
GROUP_CROPID = "group_cropid"
GROUP_POSX = "group_x"
GROUP_POSY = "group_y"
GROUP_SLIDEAC = "group_slideac"
GROUP_PANORMAID = "group_panoid"
GROUP_ACID = "group_acid"
GROUP_ROIID = "group_roiid"
GROUP_SLIDENUMBER = "group_slidenumber"
GROUP_SAMPLEBLOCKNAME = "group_sampleblockname"
GROUP_BASENAME = "group_basename"

META_REGEXP = "re_meta"
IMAGE_OME_FOLDER_DIRS = "ome_folder_dirs"
IMAGE_OME_META_REGEXP = "ome_meta_regexp"
IMAGE_SLIDE_REGEXP = "slide_regexp"
IMAGE_HEIGHT_PREFIX = "image_height_col_prefix"
IMAGE_WIDTH_PREFIX = "image_width_col_prefix"

CON_SQLITE = "sqlite"
CON_MYSQL = "mysql"
CON_POSTGRESQL = "postgresql"

LAYOUT_CSV_PLATE_NAME = "plate_col"
LAYOUT_CSV_WELL_NAME = "well_col"
LAYOUT_CSV_COND_NAME = "condition_col"
LAYOUT_CSV_CONTROL_NAME = "control_col"
LAYOUT_CSV_TIMEPOINT_NAME = "timepoint_col"
LAYOUT_CSV_BC_PLATE_NAME = "bc_plate_col"
LAYOUT_CSV_CONCENTRATION_NAME = "concentration_col"
LAYOUT_CSV_BARCODE = "barcode_col"
LAYOUT_CSV_COND_ID = "condition_id_col"

BC_CSV_PLATE_NAME = "plate_col"
BC_CSV_WELL_NAME = "well_col"

PANEL_CSV_CHANNEL_NAME = "channel_name"
PANEL_CSV_DISPLAY_NAME = "display_name"
PANEL_CSV_ILASTIK_NAME = "ilastik_name"
PANEL_CSV_BARCODE_NAME = "barcode_name"
PANEL_CSV_CLONE_NAME = "clone_name"
PANEL_CSV_CONCENTRATION_NAME = "concentration_name"
PANEL_CSV_TUBE_NAME = "tube_name"
PANEL_CSV_TARGET_NAME = "target"

COLMAP = "column_map"

"""
Default settings
"""
default_dict = {
    QUERY_DEFAULTS: {
        DEFAULT_OBJECT_TYPE: "cell",
        CHANNEL_MEASUREMENTS: {
            DEFAULT_MEASUREMENT_NAME: "MeanIntensityComp",
            DEFAULT_STACK_NAME: "FullStackFiltered",
            DEFAULT_MEASUREMENT_TYPE: "Intensity",
        },
        RAWDIST: {
            DEFAULT_MEASUREMENT_NAME: "MeanIntensity",
            DEFAULT_STACK_NAME: "DistStack",
            DEFAULT_MEASUREMENT_TYPE: "Intensity",
            DEFAULT_CHANNEL_NAME: "dist-sphere",
        },
        CORRDIST: {
            DEFAULT_STACK_NAME: "ObjectStack",
            DEFAULT_MEASUREMENT_TYPE: "Location",
            DEFAULT_MEASUREMENT_NAME: "dist-rim",
            DEFAULT_CHANNEL_NAME: "object",
            DEFAULT_CHANNEL_TYPE: "object",
        },
        OBJECT_DEFAULTS: {
            DEFAULT_STACK_NAME: "ObjectStack",
            DEFAULT_CHANNEL_NAME: "object",
            DEFAULT_CHANNEL_TYPE: "object",
        },
    },
    IMAGENUMBER: "ImageNumber",
    OBJECTNUMBER: "ObjectNumber",
    OBJECTTYPE: "ObjectID",
    STACK_RELATIONS: {
        PARENT: "Parent",
        MODNAME: "ModificationName",
        MODPRE: "ModificationPrefix",
        STACK: "Stack",
        REF: "RefStack",
    },
    STACK_DIR: {
        STACK: "StackName",
        ID: "index",
        NAME: "name",
        TYPE: "channel_type",
        SEP: ",",
    },
    LAYOUT_CSV: {
        PATH: None,
        LAYOUT_CSV_PLATE_NAME: "plate",
        LAYOUT_CSV_BC_PLATE_NAME: "bc_plate",
        LAYOUT_CSV_WELL_NAME: "TargetWell",
        LAYOUT_CSV_COND_NAME: None,
        LAYOUT_CSV_CONTROL_NAME: "control",
        LAYOUT_CSV_TIMEPOINT_NAME: None,
        LAYOUT_CSV_CONCENTRATION_NAME: "concentration",
        LAYOUT_CSV_BARCODE: "barcode",
        LAYOUT_CSV_COND_ID: "conditionID",
        SEP: ",",
    },
    BACKEND: CON_MYSQL,
    BARCODE_CSV: {
        PATH: None,
        BC_CSV_PLATE_NAME: "Plate",
        BC_CSV_WELL_NAME: "Well",
        SEP: ",",
    },
    CPOUTPUT: {
        MEASUREMENT_CSV: {SEP: ",", FILETYPE: ".csv", DEFAULT_OBJECT: "cell"},
        RELATION_CSV: {
            SEP: ",",
            OBJECTTYPE_FROM: "First Object Name",
            OBJECTTYPE_TO: "Second Object Name",
            OBJECTNUMBER_FROM: "First Object Number",
            OBJECTNUMBER_TO: "Second Object Number",
            IMAGENUMBER_FROM: "First Image Number",
            IMAGENUMBER_TO: "Second Image Number",
            RELATIONSHIP: "Relationship",
        },
        IMAGES_CSV: {
            MASK_FILENAME_PREFIX: "ObjectsFileName_",
            STACKIMG_FILENAME_PREFIX: "FileName_",
            SEP: ",",
            SCALING_PREFIX: "Scaling_",
            # in CP width and height seem to be switched
            IMAGE_HEIGHT_PREFIX: "Height_",
            IMAGE_WIDTH_PREFIX: "Width_",
            META_REGEXP: (
                "(?P<{}>.*)_l(?P<{}>[0-9]*)_x(?P<{}>[0-9]*)_y(?P<{}>[0-9]*).tiff".format(
                    "basename",
                    db.images.crop_number.key,
                    db.images.image_pos_x.key,
                    db.images.image_pos_y.key,
                )
            ),
            MASK_DIR: None,  # default take cpoutput dir
            GROUP_BASENAME: "basename",
            GROUP_CROPID: db.images.crop_number.key,
            GROUP_SITE: db.sites.site_name.key,
            GROUP_POSX: db.images.image_pos_x.key,
            GROUP_POSY: db.images.image_pos_y.key,
            IMAGE_OME_FOLDER_DIRS: [],
            IMAGE_OME_META_REGEXP: (
                "(?P<{}>.*)_s0_p(?P<{}>[0-9]+)_r(?P<{}>[0-9]+)_a(?P<{}>[0-9]+)_ac.*".format(
                    db.slideacs.slideac_name.key,
                    db.sites.site_mcd_panoramaid.key,
                    db.acquisitions.acquisition_mcd_roiid.key,
                    db.acquisitions.acquisition_mcd_acid.key,
                )
            ),
            GROUP_SLIDEAC: db.slideacs.slideac_name.key,
            GROUP_PANORMAID: db.sites.site_mcd_panoramaid.key,
            GROUP_ACID: db.acquisitions.acquisition_mcd_acid.key,
            GROUP_ROIID: db.acquisitions.acquisition_mcd_roiid.key,
            # IMAGE_SLIDE_REGEXP: '.*_(?P<{}>.*)_slide(?P<{}>[0-9]+)_'.format(
            #    db.sampleblocks.sampleblock_name.key, db.slides.slide_number.key),
            IMAGE_SLIDE_REGEXP: ".*_slide(?P<{}>[0-9]+)_".format(
                db.slides.slide_number.key
            ),
            GROUP_SLIDENUMBER: db.slides.slide_number.key,
            GROUP_SAMPLEBLOCKNAME: db.sampleblocks.sampleblock_name.key,
        },
    },
    PANNEL_CSV: {
        SEP: ",",
        PANEL_CSV_CHANNEL_NAME: "metal",
        PANEL_CSV_DISPLAY_NAME: "name",
        PANEL_CSV_ILASTIK_NAME: "ilastik",
        PANEL_CSV_BARCODE_NAME: "barcode",
        PANEL_CSV_CLONE_NAME: "Antibody Clone",
        PANEL_CSV_CONCENTRATION_NAME: "Final Concentration / Dilution",
        PANEL_CSV_TUBE_NAME: "Tube Number",
        PANEL_CSV_TARGET_NAME: "Target",
    },
}

"""
Required fields
"""
# TODO

"""
Functions
"""


def read_configuration(path):
    """
    Reads a configuration file into a dictionary.

    Args:
        path: path to a yaml configuration file

    Return:
        the loaded configuration file

    """

    with open(path, "r") as stream:
        try:
            conf = yaml.load(stream)
        except yaml.YAMLError as exc:
            print(exc)
    conf = add_defaults(conf)
    return conf


def add_defaults(config):
    """
    Adds the default argument to  the configuration dictionary.

    Args:
        config: a dictionary containing the configurations

    Return:
        the updated configuration file
    """

    new_config = _update_nested_dict(default_dict, config)

    return new_config


def _update_nested_dict(input_d, u):
    """
    Updates a nested dictionary
    From: https://stackoverflow.com/a/3233356
    Example:
        >>> import pprint
        >>> a = {'a': {'b': 1, 'c': {'d': 1, 'e': 1}}, 'f': 1}
        >>> b = {'a': {'c': {'e': 3, 'g': 2}}, 'f': 2}
        >>> a_updated = _update_nested_dict(a,b)
        >>> pprint.pprint(a_updated)
        {'a': {'b': 1, 'c': {'d': 1, 'e': 3, 'g': 2}}, 'f': 2}
    """

    d = copy.deepcopy(input_d)
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = _update_nested_dict(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d



class BaseFilter(object):
    def __init__(self, bro):
        self.bro = bro
        self.session = self.bro.data.main_session
        self.data = self.bro.data
        self.doquery = self.data.query_df
